# Copyright 2011 Omniscale (http://omniscale.com)
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from imposm.base import Relation, Way
from imposm.multipolygon import RelationBuilder, Ring, merge_rings

from nose.tools import eq_

def test_simple_polygon_w_hole():
    w1 = Way(1, {}, [1, 2, 3, 4, 1])
    w1.coords = [(0, 0), (10, 0), (10, 10), (0, 10), (0, 0)]
    w2 = Way(2, {}, [5, 6, 7, 8, 5])
    w2.coords = [(2, 2), (8, 2), (8, 8), (2, 8), (2, 2)]
    
    r = Relation(1, {}, [(1, 'way', 'outer'), (2, 'way', 'inner')])
    builder = RelationBuilder(r, None, None)
    rings = builder.build_rings([w1, w2])
    eq_(len(rings), 2)
    eq_(rings[0].geom.area, 100)
    eq_(rings[1].geom.area, 36)
    
    builder.build_relation_geometry(rings)
    
    eq_(r.geom.area, 100-36)

def test_simple_polygon_from_two_lines():
    w1 = Way(1, {}, [1, 2, 3])
    w1.coords = [(0, 0), (10, 0), (10, 10)]
    w2 = Way(2, {}, [3, 4, 1])
    w2.coords = [(10, 10), (0, 10), (0, 0)]
    
    r = Relation(1, {}, [(1, 'way', 'outer'), (2, 'way', 'inner')])
    builder = RelationBuilder(r, None, None)
    rings = builder.build_rings([w1, w2])
    eq_(len(rings), 1)
    eq_(rings[0].geom.area, 100)
    
    builder.build_relation_geometry(rings)
    
    eq_(r.geom.area, 100)


def test_merge_rings():
    w1 = Way(1, {}, [1, 2, 3])
    w1.coords = [(0, 0), (10, 0), (10, 10)]
    r1 = Ring(w1)
    w2 = Way(2, {}, [3, 4, 1])
    w2.coords = [(10, 10), (0, 10), (0, 0)]
    r2 = Ring(w2)
    
    rings = merge_rings([r1, r2])
    eq_(len(rings), 1)
    r = rings[0]
    eq_(r.is_closed(), True)
    # eq_(r.ways, [w1, w2])

def test_merge_rings_reverse_endpoint():
    w1 = Way(1, {'name': 'foo'}, [1, 2, 3, 4])
    w1.coords = []
    r1 = Ring(w1)
    w2 = Way(2, {'building': 'true'}, [6, 5, 4])
    w2.coords = []
    r2 = Ring(w2)
    w3 = Way(3, {}, [1, 7, 6])
    w3.coords = []
    r3 = Ring(w3)
    
    rings = merge_rings([r1, r2, r3])
    eq_(len(rings), 1)
    r = rings[0]
    eq_(r.tags, {'name': 'foo', 'building': 'true'})
    eq_(r.is_closed(), True)
    # eq_(r.ways, [w1, w2, w3])


class W(Way):
    # way without coords
    coords = []

from imposm.merge import permutations

def test_merge_rings_permutations():
    """
    Test all possible permutations of 4 ring segments.
    """
    for i in range(16):
        # test each segment in both directions
        f1 = i & 1 == 0
        f2 = i & 2 == 0
        f3 = i & 4 == 0
        f4 = i & 8 == 0
        for i1, i2, i3, i4 in permutations([0, 1, 2, 3]):
            ways = [
                W(1, {}, [1, 2, 3, 4] if f1 else [4, 3, 2, 1]),
                W(2, {}, [4, 5, 6, 7] if f2 else [7, 6, 5, 4]),
                W(3, {}, [7, 8, 9, 10] if f3 else [10, 9, 8, 7]),
                W(4, {}, [10, 11, 12, 1] if f4 else [1, 12, 11, 10]),
            ]
            ways = [ways[i1], ways[i2], ways[i3], ways[i4]]
            rings = [Ring(w) for w in ways]
            
            merged_rings = merge_rings(rings)
            eq_(len(merged_rings), 1)
            r = merged_rings[0]
            eq_(r.is_closed(), True, (ways, r.refs))
            eq_(set(r.ways), set(ways))

            # check order of refs
            prev_x = r.refs[0]
            for x in r.refs[1:]:
                if not abs(prev_x - x) == 1:
                    assert (
                        (prev_x == 1 and x == 12) or 
                        (prev_x == 12 and x == 1)
                    ), 'not in order %r' % r.refs
                prev_x = x


# Copyright 2011 Omniscale (http://omniscale.com)
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array

from multiprocessing import Process, JoinableQueue

from imposm.dbimporter import NodeProcess, WayProcess, RelationProcess
from imposm.util import create_pool, shutdown_pool
from imposm.cache import OSMCache

class ImposmWriter(object):
    def __init__(self, mapping, db, osm_cache=None, pool_size=2, logger=None, dry_run=False):
        self.mapping = mapping
        self.db = db
        self.mapper = mapping
        self.osm_cache = osm_cache or OSMCache('.')
        self.pool_size = pool_size
        self.logger = logger
        self.dry_run = dry_run

    def _write_elem(self, proc, elem_cache, log, pool_size, proc_args=[]):
        queue = JoinableQueue(16)

        importer = lambda: proc(queue, self.db, self.mapper, self.osm_cache, self.dry_run, *proc_args)
        pool = create_pool(importer, pool_size)

        data = []
        for i, elem in enumerate(elem_cache):
            if elem.tags:
                data.append(elem)
            if len(data) >= 128:
                queue.put(data)
                log.log(i)
                data = []
        queue.put(data)

        shutdown_pool(pool, queue)
        log.stop()
        self.osm_cache.close_all()

    def relations(self):
        cache = self.osm_cache.relations_cache()
        log = self.logger('relations', len(cache))
        inserted_way_queue = JoinableQueue()
        way_marker = WayMarkerProcess(inserted_way_queue, self.osm_cache, self.logger)
        way_marker.start()

        self._write_elem(RelationProcess, cache, log, self.pool_size, [inserted_way_queue])

        inserted_way_queue.put(None)
        way_marker.join()

    def ways(self):
        cache = self.osm_cache.ways_cache()
        log = self.logger('ways', len(cache))
        self._write_elem(WayProcess, cache, log, self.pool_size)

    def nodes(self):
        cache = self.osm_cache.nodes_cache()
        log = self.logger('nodes', len(cache))
        self._write_elem(NodeProcess, cache, log, self.pool_size)
        

class WayMarkerProcess(Process):
    def __init__(self, queue, osm_cache, logger):
        Process.__init__(self)
        self.daemon = True
        self.queue = queue
        self.osm_cache = osm_cache
        self.logger = logger
    
    def run(self):
        inserted_ways = array.array('I')
        while True:
            osmid = self.queue.get()
            if osmid is None:
                break
            inserted_ways.append(osmid)

        self.update_inserted_ways(inserted_ways)

    def update_inserted_ways(self, inserted_ways):
        log = self.logger('marking inserted ways', len(inserted_ways))
        cache = self.osm_cache.ways_cache(mode='w')
        for i, osmid in enumerate(inserted_ways):
            way = cache.get(osmid)
            way.tags['_inserted_'] = True
            cache.put(osmid, way.tags, way.refs)
            if i % 100 == 0:
                log.log(i)
        cache.close()
        log.stop()